'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var pqb = require('pqb');
var zod = require('zod');
var node_buffer = require('node:buffer');

const modelToZod = (model) => {
  return instanceToZod(new model().columns);
};
const instanceToZod = ({
  shape
}) => {
  const result = {};
  for (const key in shape) {
    result[key] = columnToZod(shape[key]);
  }
  return zod.z.object(result);
};
const columnToZod = (column) => {
  var _a;
  const dataType = ((_a = column.data.as) == null ? void 0 : _a.dataType) || column.dataType;
  const converter = converters[dataType];
  if (!converter)
    throw new Error(`Cannot parse column ${dataType}`);
  return converter(column);
};
const typeHandler = (fn) => {
  return (column) => {
    let type = fn(column);
    column.chain.forEach((item) => {
      if (item[0] === "transform") {
        type = type.transform(item[1]);
      } else if (item[0] === "to") {
        type = zod.z.preprocess(item[1], itemToZod(item[2]));
      } else if (item[0] === "refine") {
        type = type.refine(item[1]);
      } else if (item[0] === "superRefine") {
        type = type.superRefine(item[1]);
      }
    });
    if ("nullable" in column.data && column.data.nullable || column.isNullable) {
      if ("optional" in column.data && column.data.optional) {
        type = type.nullish();
      } else {
        type = type.nullable();
      }
    } else if ("optional" in column.data && column.data.optional) {
      type = type.optional();
    }
    if (column instanceof pqb.ColumnType) {
      if (column.data.validationDefault !== void 0) {
        type = type.default(column.data.validationDefault);
      }
    } else if (column.data.default !== void 0) {
      type = type.default(column.data.default);
    }
    return type;
  };
};
const stringParams = [
  "min",
  "max",
  "length",
  "regex",
  "startsWith",
  "endsWith"
];
const stringEmptyParams = ["email", "url", "uuid", "cuid", "trim"];
const handleString = typeHandler((column) => {
  let type = zod.z.string();
  stringParams.forEach((key) => {
    const value = column.data[key];
    if (value !== void 0) {
      type = type[key](value);
    }
  });
  stringEmptyParams.forEach((key) => {
    const value = column.data[key];
    if (value) {
      type = type[key]();
    }
  });
  return type;
});
const numberParams = ["lt", "lte", "gt", "gte", "multipleOf"];
const handleNumber = typeHandler((column) => {
  let type = zod.z.number();
  numberParams.forEach((key) => {
    const value = column.data[key];
    if (value !== void 0) {
      type = type[key](value);
    }
  });
  if (column.data.int) {
    type = type.int();
  }
  return type;
});
const handleBigInt = typeHandler(() => {
  return zod.z.string().refine(
    (value) => {
      try {
        BigInt(value);
        return true;
      } catch (_) {
        return false;
      }
    },
    {
      message: "Failed to parse bigint"
    }
  );
});
const handleBuffer = typeHandler(() => {
  return zod.z.instanceof(node_buffer.Buffer);
});
const dateParams = ["min", "max"];
const handleDate = typeHandler((column) => {
  let type = zod.z.date();
  dateParams.forEach((key) => {
    const value = column.data[key];
    if (value !== void 0) {
      type = type[key](value);
    }
  });
  return zod.z.preprocess(
    (val) => typeof val === "string" ? new Date(val) : val,
    type
  );
});
const handleTime = typeHandler(() => {
  return zod.z.string().refine(
    (val) => {
      return !isNaN(new Date(`2000-01-01 ${val}`).getTime());
    },
    {
      message: "Invalid time"
    }
  );
});
const interval = zod.z.object({
  years: zod.z.number().optional(),
  months: zod.z.number().optional(),
  days: zod.z.number().optional(),
  hours: zod.z.number().optional(),
  seconds: zod.z.number().optional()
}).strict();
const handleInterval = typeHandler(() => {
  return interval;
});
const handleBoolean = typeHandler(() => {
  return zod.z.boolean();
});
const handleEnum = typeHandler((column) => {
  const enumColumn = column;
  return zod.z.enum(enumColumn.options);
});
const handleBitString = typeHandler(() => {
  return zod.z.string().regex(/[10]/g);
});
const handleUUID = typeHandler(() => {
  return zod.z.string().uuid();
});
const arrayParams = ["min", "max", "length"];
const handleArray = typeHandler(
  (array) => {
    let type;
    if ("element" in array) {
      type = zod.z.array(jsonItemToZod(array.element));
    } else {
      type = zod.z.array(columnToZod(array.data.item));
    }
    arrayParams.forEach((key) => {
      const value = array.data[key];
      if (value !== void 0) {
        type = type[key](value);
      }
    });
    return type;
  }
);
const handleJson = typeHandler((column) => {
  const type = column.data.schema;
  return jsonItemToZod(type);
});
const jsonItemToZod = (type) => {
  const converter = jsonConverters[type.dataType];
  if (!converter)
    throw new Error(`Cannot parse column ${type.dataType}`);
  return converter(type);
};
const itemToZod = (item) => {
  return item instanceof pqb.ColumnType ? columnToZod(item) : jsonItemToZod(item);
};
const converters = {
  varchar: handleString,
  char: handleString,
  text: handleString,
  smallint: handleNumber,
  integer: handleNumber,
  real: handleNumber,
  smallserial: handleNumber,
  serial: handleNumber,
  money: handleNumber,
  bigint: handleBigInt,
  decimal: handleBigInt,
  "double precision": handleBigInt,
  bigserial: handleBigInt,
  bytea: handleBuffer,
  date: handleDate,
  timestamp: handleDate,
  "timestamp with time zone": handleDate,
  time: handleTime,
  "time with time zone": handleTime,
  interval: handleInterval,
  boolean: handleBoolean,
  enum: handleEnum,
  point: handleString,
  line: handleString,
  lseg: handleString,
  box: handleString,
  path: handleString,
  polygon: handleString,
  circle: handleString,
  cidr: handleString,
  inet: handleString,
  macaddr: handleString,
  macaddr8: handleString,
  bit: handleBitString,
  "bit varying": handleBitString,
  tsvector: handleString,
  tsquery: handleString,
  xml: handleString,
  json: handleString,
  uuid: handleUUID,
  array: handleArray,
  jsonb: handleJson
};
const handleAny = typeHandler(() => {
  return zod.z.any();
});
const handleNaN = typeHandler(() => {
  return zod.z.nan();
});
const handleNever = typeHandler(() => {
  return zod.z.never();
});
const handleNull = typeHandler(() => {
  return zod.z.null();
});
const handleUndefined = typeHandler(() => {
  return zod.z.undefined();
});
const handleUnknown = typeHandler(() => {
  return zod.z.unknown();
});
const handleVoid = typeHandler(() => {
  return zod.z.void();
});
const handleInstanceOf = typeHandler(
  (type) => {
    return zod.z.instanceof(type.class);
  }
);
const handleLiteral = typeHandler((type) => {
  return zod.z.literal(type.value);
});
const handleMap = typeHandler((type) => {
  const { keyType, valueType } = type;
  return zod.z.map(jsonItemToZod(keyType), jsonItemToZod(valueType));
});
const setParams = ["min", "max", "size"];
const handleSet = typeHandler((column) => {
  const { valueType } = column;
  let type = zod.z.set(jsonItemToZod(valueType));
  setParams.forEach((key) => {
    const value = column.data[key];
    if (value !== void 0) {
      type = type[key](value);
    }
  });
  return type;
});
const handleNativeEnum = typeHandler((type) => {
  return zod.z.nativeEnum(type.enum);
});
const handleTuple = typeHandler((column) => {
  let type = zod.z.tuple(
    column.items.map((item) => jsonItemToZod(item))
  );
  if (column.restType) {
    type = type.rest(jsonItemToZod(column.restType));
  }
  return type;
});
const handleObject = typeHandler(
  (type) => {
    const shape = {};
    for (const key in type.shape) {
      shape[key] = jsonItemToZod(type.shape[key]);
    }
    let object = zod.z.object(shape);
    if (type.unknownKeys === "passthrough") {
      object = object.passthrough();
    } else if (type.unknownKeys === "strict") {
      object = object.strict();
    }
    if (type.catchAllType) {
      object = object.catchall(jsonItemToZod(type.catchAllType));
    }
    return object;
  }
);
const handleRecord = typeHandler(
  (type) => {
    return zod.z.record(jsonItemToZod(type.keyType), jsonItemToZod(type.valueType));
  }
);
const handleUnion = typeHandler(
  (type) => {
    return zod.z.union(type.types.map(jsonItemToZod));
  }
);
const handleIntersection = typeHandler(
  (type) => {
    return zod.z.intersection(jsonItemToZod(type.left), jsonItemToZod(type.right));
  }
);
const handleDiscriminatedUnion = typeHandler(
  (type) => {
    return zod.z.discriminatedUnion(
      type.discriminator,
      [...type.options.values()].map(jsonItemToZod)
    );
  }
);
const handleLazy = typeHandler((type) => {
  return zod.z.lazy(() => jsonItemToZod(type.getter()));
});
const jsonConverters = {
  any: handleAny,
  bigint: handleBigInt,
  boolean: handleBoolean,
  date: handleDate,
  nan: handleNaN,
  never: handleNever,
  null: handleNull,
  number: handleNumber,
  string: handleString,
  undefined: handleUndefined,
  unknown: handleUnknown,
  void: handleVoid,
  array: handleArray,
  enum: handleEnum,
  instanceOf: handleInstanceOf,
  literal: handleLiteral,
  map: handleMap,
  set: handleSet,
  nativeEnum: handleNativeEnum,
  tuple: handleTuple,
  object: handleObject,
  record: handleRecord,
  union: handleUnion,
  intersection: handleIntersection,
  discriminatedUnion: handleDiscriminatedUnion,
  lazy: handleLazy
};

exports.columnToZod = columnToZod;
exports.instanceToZod = instanceToZod;
exports.modelToZod = modelToZod;
//# sourceMappingURL=index.js.map
